@@ -6,7 +6,7 @@ from typing import Any, Callable, Dict, Iterator, List, Optional, TypedDict, Uni
 from urllib.parse import urlunparse
 
 import requests
-from typing_extensions import NotRequired
+from typing_extensions import NotRequired, deprecated
 
 from snowflake import snowpark
 from snowflake.cortex._sse_client import SSEClient
@@ -23,6 +23,15 @@ logger = logging.getLogger(__name__)
 _REST_COMPLETE_URL = "/api/v2/cortex/inference:complete"
 
 
+class ResponseFormat(TypedDict):
+    """Represents an object describing response format config for structured-output mode"""
+
+    type: str
+    """The response format type (e.g. "json")"""
+    schema: Dict[str, Any]
+    """The schema defining the structure of the response. For json it should be a valid json schema object"""
+
+
 class ConversationMessage(TypedDict):
     """Represents an conversation interaction."""
 
@@ -49,6 +58,13 @@ class CompleteOptions(TypedDict):
     generally used as an alternative to temperature. The difference is that top_p restricts the set of possible tokens
     that the model outputs, while temperature influences which tokens are chosen at each step. """
 
+    guardrails: NotRequired[bool]
+    """ A boolean value that controls whether Cortex Guard filters unsafe or harmful responses
+    from the language model. """
+
+    response_format: NotRequired[ResponseFormat]
+    """ An object describing response format config for structured-output mode """
+
 
 class ResponseParseException(Exception):
     """This exception is raised when the server response cannot be parsed."""
@@ -56,6 +72,15 @@ class ResponseParseException(Exception):
     pass
 
 
+class GuardrailsOptions(TypedDict):
+    enabled: bool
+    """A boolean value that controls whether Cortex Guard filters unsafe or harmful responses
+    from the language model."""
+
+    response_when_unsafe: str
+    """The response to return when the language model generates unsafe or harmful content."""
+
+
 _MAX_RETRY_SECONDS = 30
 
 
@@ -95,6 +120,32 @@ def _make_common_request_headers() -> Dict[str, str]:
     return headers
 
 
+def _validate_response_format_object(options: CompleteOptions) -> None:
+    """Validate the response format object for structured-output mode.
+
+    More details can be found in:
+    docs.snowflake.com/en/user-guide/snowflake-cortex/complete-structured-outputs#using-complete-structured-outputs
+
+    Args:
+        options: The complete options object.
+
+    Raises:
+        ValueError: If the response format object is invalid or missing required fields.
+    """
+    if options is not None and options.get("response_format") is not None:
+        options_obj = options.get("response_format")
+        if not isinstance(options_obj, dict):
+            raise ValueError("'response_format' should be an object")
+        if options_obj.get("type") is None:
+            raise ValueError("'type' cannot be empty for 'response_format' object")
+        if not isinstance(options_obj.get("type"), str):
+            raise ValueError("'type' needs to be a str for 'response_format' object")
+        if options_obj.get("schema") is None:
+            raise ValueError("'schema' cannot be empty for 'response_format' object")
+        if not isinstance(options_obj.get("schema"), dict):
+            raise ValueError("'schema' needs to be a dict for 'response_format' object")
+
+
 def _make_request_body(
     model: str,
     prompt: Union[str, List[ConversationMessage]],
@@ -117,18 +168,45 @@ def _make_request_body(
             data["temperature"] = options["temperature"]
         if "top_p" in options:
             data["top_p"] = options["top_p"]
+        if "guardrails" in options and options["guardrails"]:
+            guardrails_options: GuardrailsOptions = {
+                "enabled": True,
+                "response_when_unsafe": "Response filtered by Cortex Guard",
+            }
+            data["guardrails"] = guardrails_options
+        if "response_format" in options:
+            data["response_format"] = options["response_format"]
+
     return data
 
 
 # XP endpoint returns a dict response which needs to be converted to a format which can
 # be consumed by the SSEClient. This method does that.
 def _xp_dict_to_response(raw_resp: Dict[str, Any]) -> requests.Response:
+
     response = requests.Response()
     response.status_code = int(raw_resp["status"])
     response.headers = raw_resp["headers"]
 
+    request_id = None
+    for key, value in raw_resp["headers"].items():
+        # Note: there is some whitespace in the headers making it not possible
+        # to directly index the header reliably.
+        if key.strip().lower() == "x-snowflake-request-id":
+            request_id = value
+            break
+
     data = raw_resp["content"]
-    data = json.loads(data)
+    try:
+        data = json.loads(data)
+    except json.JSONDecodeError:
+        raise ValueError(f"Request failed (request id: {request_id})")
+    if response.status_code < 200 or response.status_code >= 300:
+        if "message" not in data:
+            raise ValueError(f"Request failed (request id: {request_id})")
+        message = data["message"]
+        raise ValueError(f"Request failed: {message} (request id: {request_id})")
+
     # Convert the dictionary to a string format that resembles the SSE event format
     # For example, if the dict is {'event': 'message', 'data': 'your data'}, it should be formatted like this:
     sse_format_data = ""
@@ -144,6 +222,7 @@ def _xp_dict_to_response(raw_resp: Dict[str, Any]) -> requests.Response:
 
 @retry
 def _call_complete_xp(
+    snow_api_xp_request_handler: Optional[Callable[..., Dict[str, Any]]],
     model: str,
     prompt: Union[str, List[ConversationMessage]],
     options: Optional[CompleteOptions] = None,
@@ -151,9 +230,8 @@ def _call_complete_xp(
 ) -> requests.Response:
     headers = _make_common_request_headers()
     body = _make_request_body(model, prompt, options)
-    import _snowflake
-
-    raw_resp = _snowflake.send_snow_api_request("POST", _REST_COMPLETE_URL, {}, headers, body, {}, deadline)
+    assert snow_api_xp_request_handler is not None
+    raw_resp = snow_api_xp_request_handler("POST", _REST_COMPLETE_URL, {}, headers, body, {}, deadline)
     return _xp_dict_to_response(raw_resp)
 
 
@@ -204,11 +282,21 @@ def _return_stream_response(response: requests.Response, deadline: Optional[floa
         if deadline is not None and time.time() > deadline:
             raise TimeoutError()
         try:
-            yield json.loads(event.data)["choices"][0]["delta"]["content"]
+            parsed_resp = json.loads(event.data)
+        except json.JSONDecodeError:
+            raise ResponseParseException("Server response cannot be parsed")
+        try:
+            yield parsed_resp["choices"][0]["delta"]["content"]
         except (json.JSONDecodeError, KeyError, IndexError):
             # For the sake of evolution of the output format,
             # ignore stream messages that don't match the expected format.
-            pass
+
+            # This is the case of midstream errors which were introduced specifically for structured output.
+            # TODO: discuss during code review
+            if parsed_resp.get("error"):
+                yield json.dumps(parsed_resp)
+            else:
+                pass
 
 
 def _complete_call_sql_function_snowpark(
@@ -218,17 +306,26 @@ def _complete_call_sql_function_snowpark(
 
 
 def _complete_non_streaming_immediate(
+    snow_api_xp_request_handler: Optional[Callable[..., Dict[str, Any]]],
     model: str,
     prompt: Union[str, List[ConversationMessage]],
     options: Optional[CompleteOptions],
     session: Optional[snowpark.Session] = None,
     deadline: Optional[float] = None,
 ) -> str:
-    response = _complete_rest(model=model, prompt=prompt, options=options, session=session, deadline=deadline)
+    response = _complete_rest(
+        snow_api_xp_request_handler=snow_api_xp_request_handler,
+        model=model,
+        prompt=prompt,
+        options=options,
+        session=session,
+        deadline=deadline,
+    )
     return "".join(response)
 
 
 def _complete_non_streaming_impl(
+    snow_api_xp_request_handler: Optional[Callable[..., Dict[str, Any]]],
     function: str,
     model: Union[str, snowpark.Column],
     prompt: Union[str, List[ConversationMessage], snowpark.Column],
@@ -245,20 +342,36 @@ def _complete_non_streaming_impl(
         raise ValueError("'model' cannot be a snowpark.Column when 'prompt' is a string.")
     if isinstance(options, snowpark.Column):
         raise ValueError("'options' cannot be a snowpark.Column when 'prompt' is a string.")
+    if options and not isinstance(options, snowpark.Column):
+        _validate_response_format_object(options)
     return _complete_non_streaming_immediate(
-        model=model, prompt=prompt, options=options, session=session, deadline=deadline
+        snow_api_xp_request_handler=snow_api_xp_request_handler,
+        model=model,
+        prompt=prompt,
+        options=options,
+        session=session,
+        deadline=deadline,
     )
 
 
 def _complete_rest(
+    snow_api_xp_request_handler: Optional[Callable[..., Dict[str, Any]]],
     model: str,
     prompt: Union[str, List[ConversationMessage]],
     options: Optional[CompleteOptions] = None,
     session: Optional[snowpark.Session] = None,
     deadline: Optional[float] = None,
 ) -> Iterator[str]:
-    if is_in_stored_procedure():  # type: ignore[no-untyped-call]
-        response = _call_complete_xp(model=model, prompt=prompt, options=options, deadline=deadline)
+    if options:
+        _validate_response_format_object(options)
+    if snow_api_xp_request_handler is not None:
+        response = _call_complete_xp(
+            snow_api_xp_request_handler=snow_api_xp_request_handler,
+            model=model,
+            prompt=prompt,
+            options=options,
+            deadline=deadline,
+        )
     else:
         response = _call_complete_rest(model=model, prompt=prompt, options=options, session=session, deadline=deadline)
     assert response.status_code >= 200 and response.status_code < 300
@@ -268,10 +381,11 @@ def _complete_rest(
 def _complete_impl(
     model: Union[str, snowpark.Column],
     prompt: Union[str, List[ConversationMessage], snowpark.Column],
+    snow_api_xp_request_handler: Optional[Callable[..., Dict[str, Any]]] = None,
+    function: str = "snowflake.cortex.complete",
     options: Optional[CompleteOptions] = None,
     session: Optional[snowpark.Session] = None,
     stream: bool = False,
-    function: str = "snowflake.cortex.complete",
     timeout: Optional[float] = None,
     deadline: Optional[float] = None,
 ) -> Union[str, Iterator[str], snowpark.Column]:
@@ -284,14 +398,29 @@ def _complete_impl(
             raise ValueError("in REST mode, 'model' must be a string")
         if not isinstance(prompt, str) and not isinstance(prompt, List):
             raise ValueError("in REST mode, 'prompt' must be a string or a list of ConversationMessage")
-        return _complete_rest(model=model, prompt=prompt, options=options, session=session, deadline=deadline)
-    return _complete_non_streaming_impl(function, model, prompt, options, session, deadline)
+        return _complete_rest(
+            snow_api_xp_request_handler=snow_api_xp_request_handler,
+            model=model,
+            prompt=prompt,
+            options=options,
+            session=session,
+            deadline=deadline,
+        )
+    return _complete_non_streaming_impl(
+        snow_api_xp_request_handler=snow_api_xp_request_handler,
+        function=function,
+        model=model,
+        prompt=prompt,
+        options=options,
+        session=session,
+        deadline=deadline,
+    )
 
 
 @telemetry.send_api_usage_telemetry(
     project=CORTEX_FUNCTIONS_TELEMETRY_PROJECT,
 )
-def Complete(
+def complete(
     model: Union[str, snowpark.Column],
     prompt: Union[str, List[ConversationMessage], snowpark.Column],
     *,
@@ -319,10 +448,19 @@ def Complete(
     Returns:
         A column of string responses.
     """
+
+    # Set the XP snow api function, if available.
+    snow_api_xp_request_handler = None
+    if is_in_stored_procedure():  # type: ignore[no-untyped-call]
+        import _snowflake
+
+        snow_api_xp_request_handler = _snowflake.send_snow_api_request
+
     try:
         return _complete_impl(
             model,
             prompt,
+            snow_api_xp_request_handler=snow_api_xp_request_handler,
             options=options,
             session=session,
             stream=stream,
@@ -331,3 +469,8 @@ def Complete(
         )
     except ValueError as err:
         raise err
+
+
+Complete = deprecated("Complete() is deprecated and will be removed in a future release. Use complete() instead")(
+    telemetry.send_api_usage_telemetry(project=CORTEX_FUNCTIONS_TELEMETRY_PROJECT)(complete)
+)
