@@ -1,3 +1,4 @@
+import enum
 import json
 import os
 import pathlib
@@ -31,8 +32,15 @@ from snowflake.snowpark import dataframe, row, session
 from snowflake.snowpark._internal import utils as snowpark_utils
 
 
+# An enum class to represent Create Or Alter Model SQL command.
+class ModelAction(enum.Enum):
+    CREATE = "CREATE"
+    ALTER = "ALTER"
+
+
 class ServiceInfo(TypedDict):
     name: str
+    status: str
     inference_endpoint: Optional[str]
 
 
@@ -91,7 +99,7 @@ class ModelOperator:
             and self._model_version_client == __value._model_version_client
         )
 
-    def prepare_model_stage_path(
+    def prepare_model_temp_stage_path(
         self,
         *,
         database_name: Optional[sql_identifier.SqlIdentifier],
@@ -109,17 +117,28 @@ class ModelOperator:
         )
         return f"@{self._stage_client.fully_qualified_object_name(database_name, schema_name, stage_name)}/model"
 
-    def create_from_stage(
+    def get_model_version_stage_path(
+        self,
+        *,
+        database_name: Optional[sql_identifier.SqlIdentifier],
+        schema_name: Optional[sql_identifier.SqlIdentifier],
+        model_name: sql_identifier.SqlIdentifier,
+        version_name: sql_identifier.SqlIdentifier,
+    ) -> str:
+        return (
+            f"snow://model/{self._stage_client.fully_qualified_object_name(database_name, schema_name, model_name)}"
+            f"/versions/{version_name}/"
+        )
+
+    def get_model_action_from_model_name_and_version(
         self,
-        composed_model: model_composer.ModelComposer,
         *,
         database_name: Optional[sql_identifier.SqlIdentifier],
         schema_name: Optional[sql_identifier.SqlIdentifier],
         model_name: sql_identifier.SqlIdentifier,
         version_name: sql_identifier.SqlIdentifier,
         statement_params: Optional[Dict[str, Any]] = None,
-    ) -> None:
-        stage_path = str(composed_model.stage_path)
+    ) -> ModelAction:
         if self.validate_existence(
             database_name=database_name,
             schema_name=schema_name,
@@ -139,6 +158,79 @@ class ModelOperator:
                     f" version {version_name} already existed."
                 )
             else:
+                return ModelAction.ALTER
+        else:
+            return ModelAction.CREATE
+
+    def add_or_create_live_version(
+        self,
+        *,
+        database_name: Optional[sql_identifier.SqlIdentifier],
+        schema_name: Optional[sql_identifier.SqlIdentifier],
+        model_name: sql_identifier.SqlIdentifier,
+        version_name: sql_identifier.SqlIdentifier,
+        statement_params: Optional[Dict[str, Any]] = None,
+    ) -> None:
+        model_action = self.get_model_action_from_model_name_and_version(
+            database_name=database_name,
+            schema_name=schema_name,
+            model_name=model_name,
+            version_name=version_name,
+            statement_params=statement_params,
+        )
+        if model_action == ModelAction.CREATE:
+            self._model_version_client.create_live_version(
+                database_name=database_name,
+                schema_name=schema_name,
+                model_name=model_name,
+                version_name=version_name,
+                statement_params=statement_params,
+            )
+        elif model_action == ModelAction.ALTER:
+            self._model_version_client.add_live_version(
+                database_name=database_name,
+                schema_name=schema_name,
+                model_name=model_name,
+                version_name=version_name,
+                statement_params=statement_params,
+            )
+        else:
+            raise AssertionError(f"The model_action is {model_action}. Expected CREATE or ALTER.")
+
+    def create_from_stage(
+        self,
+        composed_model: model_composer.ModelComposer,
+        *,
+        database_name: Optional[sql_identifier.SqlIdentifier],
+        schema_name: Optional[sql_identifier.SqlIdentifier],
+        model_name: sql_identifier.SqlIdentifier,
+        version_name: sql_identifier.SqlIdentifier,
+        statement_params: Optional[Dict[str, Any]] = None,
+        use_live_commit: Optional[bool] = False,
+    ) -> None:
+
+        if use_live_commit:
+            # if the model version is live, we can only commit the version
+            self._model_version_client.commit_version(
+                database_name=database_name,
+                schema_name=schema_name,
+                model_name=model_name,
+                version_name=version_name,
+                statement_params=statement_params,
+            )
+        else:
+            stage_path = str(composed_model.stage_path)
+            # if the model version is not live,
+            # find whether the model exists and whether the version exists
+            # and then decide whether to create or alter the model
+            model_action = self.get_model_action_from_model_name_and_version(
+                database_name=database_name,
+                schema_name=schema_name,
+                model_name=model_name,
+                version_name=version_name,
+                statement_params=statement_params,
+            )
+            if model_action == ModelAction.ALTER:
                 self._model_version_client.add_version_from_stage(
                     database_name=database_name,
                     schema_name=schema_name,
@@ -147,15 +239,17 @@ class ModelOperator:
                     version_name=version_name,
                     statement_params=statement_params,
                 )
-        else:
-            self._model_version_client.create_from_stage(
-                database_name=database_name,
-                schema_name=schema_name,
-                stage_path=stage_path,
-                model_name=model_name,
-                version_name=version_name,
-                statement_params=statement_params,
-            )
+            elif model_action == ModelAction.CREATE:
+                self._model_version_client.create_from_stage(
+                    database_name=database_name,
+                    schema_name=schema_name,
+                    stage_path=stage_path,
+                    model_name=model_name,
+                    version_name=version_name,
+                    statement_params=statement_params,
+                )
+            else:
+                raise AssertionError(f"The model_action is {model_action}. Expected CREATE or ALTER.")
 
     def create_from_model_version(
         self,
@@ -550,9 +644,13 @@ class ModelOperator:
         fully_qualified_service_names = [str(service) for service in json_array if "MODEL_BUILD_" not in service]
 
         result = []
-        ingress_url: Optional[str] = None
+
         for fully_qualified_service_name in fully_qualified_service_names:
+            ingress_url: Optional[str] = None
             db, schema, service_name = sql_identifier.parse_fully_qualified_name(fully_qualified_service_name)
+            service_status, _ = self._service_client.get_service_status(
+                database_name=db, schema_name=schema, service_name=service_name, statement_params=statement_params
+            )
             for res_row in self._service_client.show_endpoints(
                 database_name=db, schema_name=schema, service_name=service_name, statement_params=statement_params
             ):
@@ -566,7 +664,11 @@ class ModelOperator:
                     )
                     if not ingress_url.endswith(ModelOperator.INGRESS_ENDPOINT_URL_SUFFIX):
                         ingress_url = None
-            result.append(ServiceInfo(name=fully_qualified_service_name, inference_endpoint=ingress_url))
+            result.append(
+                ServiceInfo(
+                    name=fully_qualified_service_name, status=service_status.value, inference_endpoint=ingress_url
+                )
+            )
 
         return result
 
@@ -687,14 +789,17 @@ class ModelOperator:
         version_name: sql_identifier.SqlIdentifier,
         statement_params: Optional[Dict[str, Any]] = None,
     ) -> type_hints.Task:
-        model_spec = self._fetch_model_spec(
+        model_version = self._model_client.show_versions(
             database_name=database_name,
             schema_name=schema_name,
             model_name=model_name,
             version_name=version_name,
+            validate_result=True,
             statement_params=statement_params,
-        )
-        task_val = model_spec.get("task", type_hints.Task.UNKNOWN.value)
+        )[0]
+
+        model_attributes = json.loads(model_version.model_attributes)
+        task_val = model_attributes.get("task", type_hints.Task.UNKNOWN.value)
         return type_hints.Task(task_val)
 
     def get_functions(
