@@ -1,22 +1,25 @@
 import os
 from typing import TYPE_CHECKING, Callable, Dict, Optional, Type, cast, final
 
-import numpy as np
 import pandas as pd
+from packaging import version
 from typing_extensions import TypeGuard, Unpack
 
 from snowflake.ml._internal import type_utils
 from snowflake.ml.model import custom_model, model_signature, type_hints as model_types
 from snowflake.ml.model._packager.model_env import model_env
 from snowflake.ml.model._packager.model_handlers import _base, _utils as handlers_utils
-from snowflake.ml.model._packager.model_handlers_migrator import base_migrator
+from snowflake.ml.model._packager.model_handlers_migrator import (
+    base_migrator,
+    tensorflow_migrator_2023_12_01,
+    tensorflow_migrator_2025_01_01,
+)
 from snowflake.ml.model._packager.model_meta import (
     model_blob_meta,
     model_meta as model_meta_api,
     model_meta_schema,
 )
 from snowflake.ml.model._signatures import (
-    numpy_handler,
     tensorflow_handler,
     utils as model_signature_utils,
 )
@@ -27,15 +30,18 @@ if TYPE_CHECKING:
 
 @final
 class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
-    """Handler for TensorFlow based model.
+    """Handler for TensorFlow based model or keras v2 model.
 
     Currently tensorflow.Module based classes are supported.
     """
 
     HANDLER_TYPE = "tensorflow"
-    HANDLER_VERSION = "2023-12-01"
-    _MIN_SNOWPARK_ML_VERSION = "1.0.12"
-    _HANDLER_MIGRATOR_PLANS: Dict[str, Type[base_migrator.BaseModelHandlerMigrator]] = {}
+    HANDLER_VERSION = "2025-03-01"
+    _MIN_SNOWPARK_ML_VERSION = "1.8.0"
+    _HANDLER_MIGRATOR_PLANS: Dict[str, Type[base_migrator.BaseModelHandlerMigrator]] = {
+        "2023-12-01": tensorflow_migrator_2023_12_01.TensorflowHandlerMigrator20231201,
+        "2025-01-01": tensorflow_migrator_2025_01_01.TensorflowHandlerMigrator20250101,
+    }
 
     MODEL_BLOB_FILE_OR_DIR = "model"
     DEFAULT_TARGET_METHODS = ["__call__"]
@@ -45,7 +51,13 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
         cls,
         model: model_types.SupportedModelType,
     ) -> TypeGuard["tensorflow.nn.Module"]:
-        return type_utils.LazyType("tensorflow.Module").isinstance(model)
+        if not type_utils.LazyType("tensorflow.Module").isinstance(model):
+            return False
+        if type_utils.LazyType("keras.Model").isinstance(model):
+            import keras
+
+            return version.parse(keras.__version__) < version.parse("3.0.0")
+        return True
 
     @classmethod
     def cast_model(
@@ -77,11 +89,18 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
 
         assert isinstance(model, tensorflow.Module)
 
-        is_keras_model = type_utils.LazyType("tensorflow.keras.Model").isinstance(model) or type_utils.LazyType(
-            "tf_keras.Model"
-        ).isinstance(model)
+        is_keras_model = type_utils.LazyType("keras.Model").isinstance(model)
+        is_tf_keras_model = type_utils.LazyType("tf_keras.Model").isinstance(model)
+        # Tensorflow and keras model save format is different.
+        # Keras v2 models are saved using keras api
+        # Tensorflow models are saved using tensorflow api
 
-        if is_keras_model:
+        if is_keras_model or is_tf_keras_model:
+            save_format = "keras_tf"
+        else:
+            save_format = "tf"
+
+        if is_keras_model or is_tf_keras_model:
             default_target_methods = ["predict"]
         else:
             default_target_methods = cls.DEFAULT_TARGET_METHODS
@@ -93,22 +112,35 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
                 default_target_methods=default_target_methods,
             )
 
+            multiple_inputs = kwargs.get("multiple_inputs", False)
+
+            if is_keras_model and len(target_methods) > 1:
+                raise ValueError("Keras model can only have one target method.")
+
             def get_prediction(
                 target_method_name: str, sample_input_data: "model_types.SupportedLocalDataType"
             ) -> model_types.SupportedLocalDataType:
-                if not tensorflow_handler.SeqOfTensorflowTensorHandler.can_handle(sample_input_data):
-                    sample_input_data = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_from_df(
-                        model_signature._convert_local_data_to_df(sample_input_data)
-                    )
+                if multiple_inputs:
+                    if not tensorflow_handler.SeqOfTensorflowTensorHandler.can_handle(sample_input_data):
+                        sample_input_data = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_from_df(
+                            model_signature._convert_local_data_to_df(sample_input_data)
+                        )
+                else:
+                    if not tensorflow_handler.TensorflowTensorHandler.can_handle(sample_input_data):
+                        sample_input_data = tensorflow_handler.TensorflowTensorHandler.convert_from_df(
+                            model_signature._convert_local_data_to_df(sample_input_data)
+                        )
 
                 target_method = getattr(model, target_method_name, None)
                 assert callable(target_method)
                 for tensor in sample_input_data:
                     tensorflow.stop_gradient(tensor)
-                predictions_df = target_method(*sample_input_data)
-
-                if isinstance(predictions_df, (tensorflow.Tensor, tensorflow.Variable, np.ndarray)):
-                    predictions_df = [predictions_df]
+                if multiple_inputs:
+                    predictions_df = target_method(*sample_input_data)
+                    if not isinstance(predictions_df, tuple):
+                        predictions_df = [predictions_df]
+                else:
+                    predictions_df = target_method(sample_input_data)
 
                 return predictions_df
 
@@ -122,31 +154,38 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
 
         model_blob_path = os.path.join(model_blobs_dir_path, name)
         os.makedirs(model_blob_path, exist_ok=True)
-        if is_keras_model:
-            tensorflow.keras.models.save_model(model, os.path.join(model_blob_path, cls.MODEL_BLOB_FILE_OR_DIR))
-            model_meta.env.include_if_absent(
-                [
-                    model_env.ModelDependency(requirement="keras<3", pip_name="keras"),
-                ],
-                check_local_version=False,
-            )
+        save_path = os.path.join(model_blob_path, cls.MODEL_BLOB_FILE_OR_DIR)
+        if save_format == "keras_tf":
+            model.save(save_path, save_format="tf")
         else:
-            tensorflow.saved_model.save(model, os.path.join(model_blob_path, cls.MODEL_BLOB_FILE_OR_DIR))
+            tensorflow.saved_model.save(
+                model,
+                save_path,
+                options=tensorflow.saved_model.SaveOptions(experimental_custom_gradients=False),
+            )
 
         base_meta = model_blob_meta.ModelBlobMeta(
             name=name,
             model_type=cls.HANDLER_TYPE,
             handler_version=cls.HANDLER_VERSION,
             path=cls.MODEL_BLOB_FILE_OR_DIR,
-            options=model_meta_schema.TensorflowModelBlobOptions(is_keras_model=is_keras_model),
+            options=model_meta_schema.TensorflowModelBlobOptions(
+                save_format=save_format, multiple_inputs=multiple_inputs
+            ),
         )
         model_meta.models[name] = base_meta
         model_meta.min_snowpark_ml_version = cls._MIN_SNOWPARK_ML_VERSION
 
+        dependencies = [
+            model_env.ModelDependency(requirement="tensorflow", pip_name="tensorflow"),
+        ]
+        if is_keras_model:
+            dependencies.append(model_env.ModelDependency(requirement="keras<=3", pip_name="keras"))
+        elif is_tf_keras_model:
+            dependencies.append(model_env.ModelDependency(requirement="tf-keras", pip_name="tf-keras"))
+
         model_meta.env.include_if_absent(
-            [
-                model_env.ModelDependency(requirement="tensorflow", pip_name="tensorflow"),
-            ],
+            dependencies,
             check_local_version=True,
         )
         model_meta.env.cuda_version = kwargs.get("cuda_version", model_env.DEFAULT_CUDA_VERSION)
@@ -159,6 +198,7 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
         model_blobs_dir_path: str,
         **kwargs: Unpack[model_types.TensorflowLoadOptions],
     ) -> "tensorflow.Module":
+        os.environ["TF_USE_LEGACY_KERAS"] = "1"
         import tensorflow
 
         model_blob_path = os.path.join(model_blobs_dir_path, name)
@@ -166,10 +206,13 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
         model_blob_metadata = model_blobs_metadata[name]
         model_blob_filename = model_blob_metadata.path
         model_blob_options = cast(model_meta_schema.TensorflowModelBlobOptions, model_blob_metadata.options)
-        if model_blob_options.get("is_keras_model", False):
-            m = tensorflow.keras.models.load_model(os.path.join(model_blob_path, model_blob_filename), compile=False)
+        load_path = os.path.join(model_blob_path, model_blob_filename)
+        save_format = model_blob_options.get("save_format", "keras_tf")
+        if save_format == "keras_tf":
+            m = tensorflow.keras.models.load_model(load_path)
         else:
-            m = tensorflow.saved_model.load(os.path.join(model_blob_path, model_blob_filename))
+            m = tensorflow.saved_model.load(load_path)
+
         return cast(tensorflow.Module, m)
 
     @classmethod
@@ -188,6 +231,10 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
             raw_model: "tensorflow.Module",
             model_meta: model_meta_api.ModelMetadata,
         ) -> Type[custom_model.CustomModel]:
+            multiple_inputs = cast(
+                model_meta_schema.TensorflowModelBlobOptions, model_meta.models[model_meta.name].options
+            )["multiple_inputs"]
+
             def fn_factory(
                 raw_model: "tensorflow.Module",
                 signature: model_signature.ModelSignature,
@@ -198,21 +245,25 @@ class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
                     if X.isnull().any(axis=None):
                         raise ValueError("Tensor cannot handle null values.")
 
-                    t = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_from_df(X, signature.inputs)
+                    if multiple_inputs:
+                        t = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_from_df(X, signature.inputs)
 
-                    for tensor in t:
-                        tensorflow.stop_gradient(tensor)
-                    res = getattr(raw_model, target_method)(*t)
+                        for tensor in t:
+                            tensorflow.stop_gradient(tensor)
+                        res = getattr(raw_model, target_method)(*t)
 
-                    if isinstance(res, (tensorflow.Tensor, tensorflow.Variable, np.ndarray)):
-                        res = [res]
-
-                    if isinstance(res, list) and len(res) > 0 and isinstance(res[0], np.ndarray):
-                        # In case of running on CPU, it will return numpy array
-                        df = numpy_handler.SeqOfNumpyArrayHandler.convert_to_df(res)
+                        if not isinstance(res, tuple):
+                            res = [res]
                     else:
-                        df = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_to_df(res)
-                    return model_signature_utils.rename_pandas_df(df, signature.outputs)
+                        t = tensorflow_handler.TensorflowTensorHandler.convert_from_df(X, signature.inputs)
+
+                        tensorflow.stop_gradient(t)
+                        res = getattr(raw_model, target_method)(t)
+
+                    return model_signature_utils.rename_pandas_df(
+                        model_signature._convert_local_data_to_df(res, ensure_serializable=True),
+                        features=signature.outputs,
+                    )
 
                 return fn
 
